
#include "TCPMobileDeviceApp.h"
#include "IPAddressResolver.h"

#define MSGKIND_CONNECT  0
#define MSGKIND_SEND     1
#define MSGKIND_MOVE     2

Define_Module(TCPMobileDeviceApp);

// Constructs
TCPMobileDeviceApp::TCPMobileDeviceApp() {
	connectMsg=NULL;
	sendMsg=NULL;
	moveMsg=NULL;
}

// Destructs
TCPMobileDeviceApp::~TCPMobileDeviceApp() {
	cancelAndDelete(connectMsg);
	cancelAndDelete(sendMsg);
	cancelAndDelete(moveMsg);
}

// Application init
void TCPMobileDeviceApp::initialize() {
    numSessions=0;
    packetsSent=0;
    bytesSent=0;
    WATCH(numSessions);
    WATCH(packetsSent);
    WATCH(bytesSent);
    connectMsg=new cMessage("connect");
    sendMsg=new cMessage("send");
    moveMsg=new cMessage("move");

	// Socket init
    const char *address=par("address");
    int port=par("port");
    socket.bind(*address?IPvXAddress(address):IPvXAddress(), port);
    socket.setCallbackObject(this);
    socket.setOutputGate(gate("tcpOut"));

	// Schedule connection
    connectMsg->setKind(MSGKIND_CONNECT);
    scheduleAt((simtime_t)par("startTime"), connectMsg);

	// Schedule movement
    moveMsg->setKind(MSGKIND_MOVE);
    scheduleAt(0, moveMsg);
}

// Handles incoming data or self message
void TCPMobileDeviceApp::handleMessage(cMessage *msg) {
    if (msg->isSelfMessage())
        handleTimer(msg);
    else
        socket.processMessage(msg);
}

// Handles self message for connection or data send
void TCPMobileDeviceApp::handleTimer(cMessage *msg) {
	switch (msg->kind()) {
		case MSGKIND_CONNECT:
            connect();
            break;
        case MSGKIND_SEND:
           sendData();
           break;
        case MSGKIND_MOVE:
           setErrorRate();
           break;
    }
}

// Connects to the server
void TCPMobileDeviceApp::connect() {
    // Renew socket if this is not the first connection
    socket.renewSocket();

    ev << "Connecting..." << endl;
    const char *connectAddress=par("connectAddress");
    int connectPort=par("connectPort");
    socket.connect(IPAddressResolver().resolve(connectAddress), connectPort);

    numSessions++;
}

// Close socket
void TCPMobileDeviceApp::close() {
    ev << "Closing socket..." << endl;
    socket.close();
}

// Sends data to the server
void TCPMobileDeviceApp::sendData() {
	long numBytes=par("numBytes");
	if (numBytes<1) numBytes=1;

    ev << "Sending " << numBytes << " B." << endl;
	cMessage *msg=new cMessage("data");
    msg->setByteLength(numBytes);
    socket.send(msg);

    packetsSent++;
    bytesSent+=numBytes;

	// Set GUI tag data
	setTag();

	// Schedule next data send
	sendMsg->setKind(MSGKIND_SEND);
	scheduleAt(simTime()+(simtime_t)par("idleTime"), sendMsg);
}

// Data arriving is not implemented (don't need it)
void TCPMobileDeviceApp::socketDataArrived(int, void *, cMessage *msg, bool) {
	delete msg;
}

// Sends first packet of data
void TCPMobileDeviceApp::socketEstablished(int, void *) {
	ev << "Connected. Sending first packet..." << endl;
	sendData();
}

// Closes the connection
void TCPMobileDeviceApp::socketPeerClosed(int, void *) {
    // Close the connection (if not already closed)
    if (socket.state()==TCPSocket::PEER_CLOSED) {
        ev << "Server TCP closed, closing here as well..." << endl;
        close();
    }
}

// Tries to reconnect
void TCPMobileDeviceApp::socketClosed(int, void *) {
    ev << "Connection closed." << endl;

	if (connectMsg->isScheduled()) cancelEvent(connectMsg);
	if (sendMsg->isScheduled()) cancelEvent(sendMsg);

    // Schedule reconnection
    connectMsg->setKind(MSGKIND_CONNECT);
    scheduleAt(simTime()+(simtime_t)par("reconnectTime"), connectMsg);
}

// Tries to reconnect
void TCPMobileDeviceApp::socketFailure(int, void *, int code) {
    ev << "Connection broken. Reconnecting..." << endl;

	if (connectMsg->isScheduled()) cancelEvent(connectMsg);
	if (sendMsg->isScheduled()) cancelEvent(sendMsg);

    // Schedule reconnection
    connectMsg->setKind(MSGKIND_CONNECT);
    scheduleAt(simTime()+(simtime_t)par("reconnectTime"), connectMsg);
}

// Sets GUI tag data
void TCPMobileDeviceApp::setTag() {
	if (!ev.isGUI()) return;
	char buf[64];
	sprintf(buf, "sent: %ld B\n%d packets", bytesSent, packetsSent);
	displayString().setTagArg("t", 0, buf);
}

// Records data
void TCPMobileDeviceApp::finish() {
	recordScalar("number of sessions", numSessions);
	recordScalar("packets sent", packetsSent);
	recordScalar("bytes sent", bytesSent);
}

//-----------------------------------------------

// Sets the channel error rate
void TCPMobileDeviceApp::setErrorRate() {
	int fullSignalPercent=par("fullSignalPercent");
	int noSignalPercent=par("noSignalPercent");

	// Calculate signal quality
	int randPercent=intrand(100)+1;
	int signalQuality=100; // default is perfect quality
	if (randPercent>fullSignalPercent) {
		signalQuality=0; // new default is no signal
		if (randPercent>fullSignalPercent+noSignalPercent) {
			signalQuality=intrand(99)+1; // intermediate signal
		}
	}
	ev << parentModule()->fullName() << " moving to an area with " << signalQuality << "% signal quality." << endl;

	// Calculate bit error rate
	double minError=par("minError");
	double error=pow(2, -(double)signalQuality-minError)+minError;
	if (error>1) error=1; // because of the rounding error

	// Set channel error
    cGate *outgate = parentModule()->gate("out");
    ev << "Old error rate: " << outgate->channel()->info() << endl;
    cBasicChannel *chan = check_and_cast<cBasicChannel *>(outgate->channel());
    chan->setError(error);
    ev << "New error rate: " << outgate->channel()->info() << endl;

	// Calculate the time duration of the mobile device staying in this area
	int areaTimeMin=par("areaTimeMin");
	int areaTimeMax=par("areaTimeMax");
	int areaTimeDif=areaTimeMax-areaTimeMin;
	int areaTime=areaTimeMin+intrand(areaTimeDif+1);
    ev << "Time to stay in this area: " << areaTime << " s." << endl;

	// Next call of this method in # seconds
	int nextMove=areaTime;

	// Schedule next move time
	moveMsg->setKind(MSGKIND_MOVE);
	scheduleAt(simTime()+nextMove, moveMsg);
}
