
#include "TCPServerStatApp.h"
#include "TCPSocket.h"
#include "IPvXAddress.h"
#include "TCPCommand_m.h"

// Statictics vector init
StatEntry::StatEntry() {
	connId=-1;
	numSessions=0;
	numMsgs=0;
	bytesRcvd=0;
    lastTime=0;
    totalTime=0;
    aveTime=0;
    maxTime=0;
}

// Statistics vector output
std::string StatEntry::info() const {
    std::stringstream out;
    out << "deviceIP:"; if (deviceIP.isUnspecified()) out << "?  "; else out << deviceIP << "  ";
    out << "numSessions:" << numSessions << " ";
    out << "bytesRcvd:" << bytesRcvd << " ";
    out << "numMsgs:" << numMsgs << " ";
    out << "aveTime:" << aveTime << " ";
    out << "maxTime:" << maxTime;
    return out.str();
}

// Statistics vector output
std::string StatEntry::detailedInfo() const {
    std::stringstream out;
    out << "deviceIP:"; if (deviceIP.isUnspecified()) out << "?  "; else out << deviceIP << "  ";
    out << "numSessions:" << numSessions << " ";
    out << "bytesRcvd:" << bytesRcvd << " ";
    out << "numMsgs:" << numMsgs << " ";
    out << "lastTime:" << lastTime << " ";
    out << "totalTime:" << totalTime << " ";
    out << "aveTime:" << aveTime << " ";
    out << "maxTime:" << maxTime;
    return out.str();
}

//==============================================================

Define_Module(TCPServerStatApp);

// The output method for statistics vector
std::ostream& operator<<(std::ostream& os, const StatEntry& e) {
    os << e.info();
    return os;
};

// Application init
void TCPServerStatApp::initialize() {
    const char *address=par("address");
    int port=par("port");

    bytesRcvd=0;
    WATCH(bytesRcvd);
    packetsRcvd=0;
    WATCH(packetsRcvd);

    WATCH_PTRVECTOR(stats);

    TCPSocket socket;
    socket.setOutputGate(gate("tcpOut"));
    socket.bind(address[0]?IPvXAddress(address):IPvXAddress(), port);
    socket.listen();
}

// Recognizes the message and send it to an appropriate method for further handling
void TCPServerStatApp::handleMessage(cMessage *msg) {
	TCPCommand *commandInfo;
	TCPConnectInfo *connectInfo;
	switch (msg->kind()) {
		case TCP_I_DATA:
		case TCP_I_URGENT_DATA:
			commandInfo=dynamic_cast<TCPCommand *>(msg->controlInfo());
			socketDataArrived(commandInfo->connId(), msg);
			delete msg;
			break;
		case TCP_I_ESTABLISHED:
			connectInfo=dynamic_cast<TCPConnectInfo *>(msg->controlInfo());
			socketEstablished(connectInfo->connId(), connectInfo->remoteAddr());
			delete msg;
			break;
		case TCP_I_PEER_CLOSED:
			socketPeerClosed(msg);
			break;
	}
}

// Handles the ordinary data
void TCPServerStatApp::socketDataArrived(int connId, cMessage *msg) {
	bytesRcvd+=msg->byteLength();
	packetsRcvd++;

	// Update the statistics for this mobile device (connection ID)
	for (unsigned int i=0; i<stats.size(); i++) {
		StatEntry *se=stats.at(i);
		if (se->connId==connId) {
			updateStats(se, msg);
			break;
		}
	}

	// Update the GUI tag data
	updateTag();
}

// Handles the new mobile device connection
void TCPServerStatApp::socketEstablished(int connId, IPvXAddress ip) {
	// Check if the mobile device is just reconnecting
	for (unsigned int i=0; i<stats.size(); i++) {
		StatEntry *se=stats.at(i);
		if (se->deviceIP.equals(ip)) {
			se->connId=connId;
			se->numSessions++;
			ev << "Device " << ip.str() << " reconnected (connection ID: " << connId << ")." << endl;
			return;
		}
	}

	// Create new statistics data for the new mobile device
	StatEntry *se=new StatEntry();
	se->deviceIP=ip;
	se->numSessions++;
	se->connId=connId;
	stats.push_back(se);
	ev << "Device " << ip.str() << " connected (connection ID: " << connId << ")." << endl;
}

// Handles the disconnection
void TCPServerStatApp::socketPeerClosed(cMessage *msg) {
        // Close the peer
        msg->setKind(TCP_C_CLOSE);
        send(msg, "tcpOut");
}

// Updates the statistics and time counters
void TCPServerStatApp::updateStats(StatEntry *se, cMessage *msg) {
		se->bytesRcvd+=msg->byteLength();
		se->numMsgs++;
		if (se->lastTime>0) {
	        simtime_t timeDiff=simTime()-se->lastTime;
	        se->totalTime+=timeDiff;
	        se->aveTime=se->totalTime/(se->numMsgs-1);
	        if (timeDiff>se->maxTime) se->maxTime=timeDiff;
		}
        se->lastTime=simTime();
		ev << "Statistics: " << se->detailedInfo() << endl;
}

// Updates the GUI tag data
void TCPServerStatApp::updateTag() {
	if (!ev.isGUI()) return;
	char buf[64];
	sprintf(buf, "rcvd: %ld B\n%d packets", bytesRcvd, packetsRcvd);
	displayString().setTagArg("t", 0, buf);
}

// Records data
void TCPServerStatApp::finish() {
    recordScalar("bytesRcvd", bytesRcvd);
    recordScalar("packetsRcvd", packetsRcvd);
}
